@@ -10,7 +10,7 @@ import six
 
 from .exceptions import NotTicketFolderException
 from .ticketfolder import TicketFolder
-from .utils import get_jira
+from .utils import lazy_get_jira
 
 
 logger = logging.getLogger(__name__)
@@ -41,6 +41,24 @@ def sync(args, jira, path, **kwargs):
     folder.sync()
 
 
+@command('Fetch remote changes', try_subfolders=True)
+def fetch(args, jira, path, **kwargs):
+    parser = argparse.ArgumentParser()
+    parser.parse_args(args)
+
+    folder = TicketFolder(path, jira)
+    folder.fetch()
+
+
+@command('Merge remote changes into your local copy', try_subfolders=True)
+def merge(args, jira, path, **kwargs):
+    parser = argparse.ArgumentParser()
+    parser.parse_args(args)
+
+    folder = TicketFolder(path, jira)
+    folder.merge()
+
+
 @command('Fetch and apply remote changes locally', try_subfolders=True)
 def pull(args, jira, path, **kwargs):
     parser = argparse.ArgumentParser()
@@ -107,9 +125,17 @@ def status(args, jira, path, **kwargs):
     if args.format == 'json':
         print(json.dumps(folder.status()))
     else:
+        print(
+            "# On ticket {ticket} ({url})".format(
+                ticket=folder.ticket_number,
+                url=folder.cached_issue.permalink(),
+            )
+        )
+        printed_changes = False
         for k, v in folder.status().items():
             human_heading = human_readable.get(k, k)  # Default to key name
             if v:
+                printed_changes = True
                 if isinstance(v, six.string_types):
                     if not v:
                         continue
@@ -118,6 +144,8 @@ def status(args, jira, path, **kwargs):
                 for item in v:
                     print('\t{item}'.format(item=item))
                 print('')
+        if not printed_changes:
+            print('No changes found')
 
 
 @command(
@@ -134,8 +162,7 @@ def clone(args, jira, path, **kwargs):
     args = parser.parse_args(args)
     ticket_number = args.ticket[0].upper()
 
-    folder = TicketFolder.create_ticket_folder(ticket_number, jira)
-    folder.sync()
+    TicketFolder.clone(ticket_number, jira)
 
 
 @command('Open this ticket in JIRA', try_subfolders=True)
@@ -145,7 +172,7 @@ def open(args, jira, path, **kwargs):
 
     folder = TicketFolder(path, jira)
 
-    webbrowser.open(folder.issue.permalink())
+    webbrowser.open(folder.cached_issue.permalink())
 
 
 @command('Show local issue changes')
@@ -179,7 +206,7 @@ def main():
         command_name,
         extra
     )
-    jira = get_jira()
+    jira = lazy_get_jira()
     try:
         fn(extra, jira=jira, path=os.getcwd())
     except NotTicketFolderException:
