@@ -17,7 +17,10 @@ CONFIG_JIRA = 'jira'
 NO_DETAIL_FIELDS = [
     'comment',
     'watches',
+    'attachment'
 ]
 FILE_FIELDS = [
     'description',
 ]
+
+CURRENT_REPO_VERSION = 4
