@@ -6,9 +6,11 @@ import os
 import re
 import subprocess
 
+from jira.resources import Issue
 import six
 
 from . import constants
+from . import migrations
 from .exceptions import (
     CannotInferTicketNumberFromFolderName,
     NotTicketFolderException
@@ -19,11 +21,11 @@ logger = logging.getLogger(__name__)
 
 
 class TicketFolder(object):
-    def __init__(self, path, jira):
+    def __init__(self, path, jira, migrate=True):
         self.path = os.path.realpath(
             os.path.expanduser(path)
         )
-        self.jira = jira
+        self.get_jira = jira
 
         if not os.path.isdir(self.metadata_dir):
             raise NotTicketFolderException(
@@ -33,6 +35,67 @@ class TicketFolder(object):
             )
 
         self.ticket_number = self.infer_ticket_number()
+        if migrate:
+            self.run_migrations()
+
+        # If no `new_comment.jira.txt` file exists, let's create one
+        comment_path = self.get_local_path(constants.TICKET_NEW_COMMENT)
+        if not os.path.exists(comment_path):
+            with open(comment_path, 'w') as out:
+                out.write('')
+
+    @property
+    def jira(self):
+        if not hasattr(self, '_jira'):
+            self._jira = self.get_jira()
+        return self._jira
+
+    @property
+    def issue(self):
+        if not hasattr(self, '_issue'):
+            self._issue = self.jira.issue(self.ticket_number)
+        return self._issue
+
+    def store_cached_issue(self):
+        storable = {
+            'options': self.issue._options,
+            'raw': self.issue.raw
+        }
+        with open(self.get_metadata_path('issue.json'), 'w') as out:
+            out.write(json.dumps(storable))
+
+    @property
+    def cached_issue(self):
+        if not hasattr(self, '_cached_issue'):
+            try:
+                issue_path = self.get_metadata_path('issue.json')
+                with open(issue_path, 'r') as _in:
+                    storable = json.loads(_in.read())
+                    self._cached_issue = Issue(
+                        storable['options'],
+                        None,
+                        storable['raw'],
+                    )
+            except IOError:
+                self.log(
+                    'Error encountered while loading cached issue!',
+                    level=logging.ERROR,
+                )
+                self._cached_issue = self.issue
+        return self._cached_issue
+
+    @property
+    def metadata_dir(self):
+        return os.path.join(
+            self.path,
+            constants.METADATA_DIR,
+        )
+
+    @property
+    def git_merge_base(self):
+        return self.run_git_command(
+            'merge-base', 'master', 'jira',
+        )
 
     def infer_ticket_number(self):
         raw_number = self.path.split('/')[-1:][0].upper()
@@ -45,25 +108,6 @@ class TicketFolder(object):
             )
         return raw_number
 
-    @property
-    def issue(self):
-        if not hasattr(self, '_issue'):
-            self._issue = self.jira.issue(self.ticket_number)
-        return self._issue
-
-    @property
-    def metadata_dir(self):
-        return os.path.join(
-            self.path,
-            constants.METADATA_DIR,
-        )
-
-    @property
-    def git_head(self):
-        return self.run_git_command(
-            'rev-parse', 'HEAD'
-        )
-
     def get_metadata_path(self, filename):
         return os.path.join(
             self.metadata_dir,
@@ -71,14 +115,16 @@ class TicketFolder(object):
         )
 
     def get_remote_file_metadata(self):
+        remote_files = self.get_shadow_path('.jirafs/remote_files.json')
         try:
-            with open(self.get_metadata_path('remote_files.json'), 'r') as _in:
+            with open(remote_files, 'r') as _in:
                 return json.loads(_in.read())
         except IOError:
             return {}
 
     def set_remote_file_metadata(self, data):
-        with open(self.get_metadata_path('remote_files.json'), 'w') as out:
+        remote_files = self.get_shadow_path('.jirafs/remote_files.json')
+        with open(remote_files, 'w') as out:
             out.write(
                 json.dumps(data)
             )
@@ -89,6 +135,20 @@ class TicketFolder(object):
             filename
         )
 
+    def get_shadow_path(self, filename):
+        return os.path.join(
+            self.get_metadata_path('shadow'),
+            filename,
+        )
+
+    @property
+    def version(self):
+        try:
+            with open(self.get_metadata_path('version'), 'r') as _in:
+                return int(_in.read().strip())
+        except IOError:
+            return 1
+
     @property
     def log_path(self):
         return self.get_metadata_path(constants.TICKET_OPERATION_LOG)
@@ -131,7 +191,7 @@ class TicketFolder(object):
             excludes_path
         ))
 
-        instance = cls(path, jira)
+        instance = cls(path, jira, migrate=False)
         instance.log(
             'Ticket folder for issue %s created at %s',
             (instance.ticket_number, instance.path, )
@@ -139,27 +199,41 @@ class TicketFolder(object):
         instance.run_git_command(
             'commit', '--allow-empty', '-m', 'Initialized'
         )
+        instance.run_migrations(silent=True)
+
         comment_path = instance.get_local_path(constants.TICKET_NEW_COMMENT)
         with open(comment_path, 'w') as out:
             out.write('')
+
         return instance
 
     @classmethod
-    def create_ticket_folder(cls, path, jira):
+    def clone(cls, path, jira):
         path = os.path.realpath(path)
         os.mkdir(path)
         folder = cls.initialize_ticket_folder(path, jira)
+        folder.sync()
         return folder
 
     def run_git_command(self, *args, **kwargs):
         failure_ok = kwargs.get('failure_ok', False)
+        shadow = kwargs.get('shadow', False)
+
+        if not shadow:
+            work_tree = self.path,
+            git_dir = self.get_metadata_path('git')
+        else:
+            work_tree = self.get_metadata_path('shadow')
+            git_dir = self.get_metadata_path('shadow/.git')
+
         cmd = [
             'git',
-            '--work-tree=%s' % self.path,
-            '--git-dir=%s' % self.get_metadata_path('git'),
+            '--work-tree=%s' % work_tree,
+            '--git-dir=%s' % git_dir,
         ]
         cmd.extend(args)
-        self.log('Executing git command %s', cmd, logging.DEBUG)
+
+        self.log('Executing git command %s', (cmd, ), logging.DEBUG)
         try:
             return subprocess.check_output(
                 cmd,
@@ -334,7 +408,7 @@ class TicketFolder(object):
         values = self.get_field_data_from_string(
             self.get_local_file_at_revision(
                 constants.TICKET_DETAILS,
-                self.git_head
+                self.git_merge_base,
             )
         )
         for field in constants.FILE_FIELDS:
@@ -344,7 +418,7 @@ class TicketFolder(object):
             try:
                 values[field] = self.get_local_file_at_revision(
                     file_field_path,
-                    self.git_head,
+                    self.git_merge_base,
                     failure_ok=False
                 ).strip()
             except subprocess.CalledProcessError:
@@ -372,26 +446,6 @@ class TicketFolder(object):
 
         return differing
 
-    def get_remote_differing_fields(self):
-        original_values = self.get_original_values()
-
-        differing = {}
-        for k in self.issue.raw['fields'].keys():
-            v = getattr(self.issue.fields, k)
-            if isinstance(v, six.string_types):
-                v = v.replace('\r\n', '\n').strip()
-            elif v is None:
-                v = ''
-            elif k in constants.NO_DETAIL_FIELDS:
-                continue
-            if original_values.get(k, '') != six.text_type(v).strip():
-                differing[k] = (
-                    original_values.get(k, ''),
-                    six.text_type(v).strip()
-                )
-
-        return differing
-
     def get_new_comment(self, clear=False):
         try:
             with open(
@@ -406,16 +460,14 @@ class TicketFolder(object):
         except IOError:
             return ''
 
-    def pull(self, status=None):
-        if status is None:
-            status = self.status()
-
+    def fetch(self):
         file_meta = self.get_remote_file_metadata()
 
-        for filename in status['to_download']:
+        for filename in self.get_remotely_changed():
             for attachment in self.issue.fields.attachment:
                 if attachment.filename == filename:
-                    with open(self.get_local_path(filename), 'wb') as download:
+                    shadow_filename = self.get_shadow_path(filename)
+                    with open(shadow_filename, 'wb') as download:
                         self.log(
                             'Download file "%s"',
                             (attachment.filename, ),
@@ -425,13 +477,16 @@ class TicketFolder(object):
 
         self.set_remote_file_metadata(file_meta)
 
-        values = self.get_original_values()
+        with open(self.get_shadow_path(constants.TICKET_DETAILS), 'w') as dets:
+            for field in sorted(self.issue.raw['fields'].keys()):
+                value = getattr(self.issue.fields, field)
+                if isinstance(value, six.string_types):
+                    value = value.replace('\r\n', '\n').strip()
+                elif value is None:
+                    value = ''
+                elif field in constants.NO_DETAIL_FIELDS:
+                    continue
 
-        for field, diff_values in status['remote_differs'].items():
-            values[field] = diff_values[1]
-
-        with open(self.get_local_path(constants.TICKET_DETAILS), 'w') as dets:
-            for field, value in sorted(six.iteritems(values)):
                 if not isinstance(value, six.string_types):
                     value = six.text_type(value)
 
@@ -439,7 +494,7 @@ class TicketFolder(object):
                     # Write specific fields to their own files without
                     # significant alteration
 
-                    file_field_path = self.get_local_path(
+                    file_field_path = self.get_shadow_path(
                         constants.TICKET_FILE_FIELD_TEMPLATE
                     ).format(field_name=field)
                     with open(file_field_path, 'w') as file_field_file:
@@ -458,7 +513,8 @@ class TicketFolder(object):
                         dets.write('    %s\n' % line)
                     dets.write('\n')
 
-        with open(self.get_local_path(constants.TICKET_COMMENTS), 'w') as comm:
+        comments_filename = self.get_shadow_path(constants.TICKET_COMMENTS)
+        with open(comments_filename, 'w') as comm:
             for comment in self.issue.fields.comment.comments:
                 comm.write('%s: %s::\n\n' % (comment.created, comment.author))
                 lines = comment.body.replace('\r\n', '\n').split('\n')
@@ -466,20 +522,26 @@ class TicketFolder(object):
                     comm.write('    %s\n' % line)
                 comm.write('\n')
 
-        # If no `new_comment.jira.txt` file exists, let's create one
-        comment_path = self.get_local_path(constants.TICKET_NEW_COMMENT)
-        if not os.path.exists(comment_path):
-            with open(comment_path, 'w') as out:
-                out.write('')
+        self.store_cached_issue()
 
-        self.run_git_command('add', '-A')
+        self.run_git_command('add', '-A', shadow=True)
         self.run_git_command(
-            'commit', '-m', 'Pulled remote changes', failure_ok=True
+            'commit', '-m', 'Pulled remote changes',
+            failure_ok=True, shadow=True
         )
+        self.run_git_command('push', 'origin', 'jira', shadow=True)
 
-    def push(self, status=None):
-        if status is None:
-            status = self.status()
+    def merge(self):
+        self.run_git_command('stash', '--include-untracked', failure_ok=True)
+        self.run_git_command('merge', 'jira')
+        self.run_git_command('stash', 'pop', failure_ok=True)
+
+    def pull(self):
+        self.fetch()
+        self.merge()
+
+    def push(self):
+        status = self.status()
 
         file_meta = self.get_remote_file_metadata()
 
@@ -499,8 +561,6 @@ class TicketFolder(object):
                 )
                 file_meta[filename] = attachment.created
 
-        self.set_remote_file_metadata(file_meta)
-
         comment = self.get_new_comment(clear=True)
         if comment:
             self.log('Adding comment "%s"' % comment)
@@ -520,30 +580,55 @@ class TicketFolder(object):
             )
             self.issue.update(**collected_updates)
 
+        # Commit local copy
         self.run_git_command('add', '-A', failure_ok=True)
         self.run_git_command(
             'commit', '-m', 'Pushed local changes', failure_ok=True
         )
 
+        # Commit changes to remote copy, too, so we record remote
+        # file metadata.
+        self.run_git_command('fetch', shadow=True)
+        self.run_git_command('merge', 'master')
+        self.set_remote_file_metadata(file_meta)
+        self.run_git_command('add', '-A', shadow=True)
+        self.run_git_command(
+            'commit', '-m', 'Pulled remote changes',
+            failure_ok=True, shadow=True
+        )
+        self.run_git_command('push', 'origin', 'jira', shadow=True)
+
     def sync(self):
-        status = self.status()
-        self.push(status=status)
-        self.pull(status=status)
+        self.pull()
+        self.push()
 
     def status(self):
         locally_changed = self.get_locally_changed()
-        remotely_changed = self.get_remotely_changed()
 
         status = {
             'to_upload': locally_changed,
-            'to_download': remotely_changed,
             'local_differs': self.get_local_differing_fields(),
-            'remote_differs': self.get_remote_differing_fields(),
             'new_comment': self.get_new_comment()
         }
 
         return status
 
+    def run_migrations(self, silent=False):
+        loglevel = logging.INFO
+        if silent:
+            loglevel = logging.DEBUG
+        while self.version < constants.CURRENT_REPO_VERSION:
+            migrator = getattr(
+                migrations,
+                'migration_%s' % str(self.version + 1).zfill(4)
+            )
+            self.migrate(migrator, loglevel=loglevel)
+
+    def migrate(self, migrator, loglevel=logging.INFO):
+        self.log('%s: Migration started', (migrator.__name__, ), loglevel)
+        migrator(self)
+        self.log('%s: Migration finished', (migrator.__name__, ), loglevel)
+
     def log(self, message, args=None, level=logging.INFO):
         if args is None:
             args = []
